/***************************************************************************\
 * ChapterXtractor by Christophe.Paris@free.fr                             *
 ***************************************************************************
 *  This program is free software; you can redistribute it and/or modify   *
 *  it under the terms of the GNU General Public License as published by   *
 *  the Free Software Foundation; either version 2 of the License, or      *
 *  (at your option) any later version.                                    *
 *                                                                         *
 *  This program is distributed in the hope that it will be useful,        *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *  GNU General Public License for more details.                           *
\***************************************************************************/

//---------------------------------------------------------------------------
// TODO : add possibilty to use up to 3 AVI files.
// TODO : add better SVCD support (enter length of each CD & calculate chapters)
// TODO : check frame detection with AVI longer than 1Go

// Added : ChapterXtractor ifo_file_name [dest_text_file -p0..10 -t1..99]
//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#define COL_VOBID   0
#define COL_CELLID  1
#define COL_LEN     2
#define COL_TS      3
#define COL_ANGLE   4
#define COL_LINK    5
#define COL_CHAPT   6

#include <vfw.h>
#include <IniFiles.hpp>
#include <stdio.h>

#include "main.h"
#include "About.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TMainForm *MainForm;
int AppliResult;

//---------------------------------------------------------------------------
__fastcall TMainForm::TMainForm(TComponent * Owner):TForm(Owner)
{
	MainForm->Caption = Application->Title;
	Application->HintPause = 100;
	Application->HintHidePause = 30000;
	
	AppsPath = IncludeTrailingBackslash(ExtractFilePath(Application->ExeName));
	AnsiString ReadmeFileName = AppsPath + "Readme.txt";

	if(FileExists(ReadmeFileName))
		ReadmeMemo->Lines->LoadFromFile(ReadmeFileName);
	else
		ReadmeMemo->Text = "Can't find Readme.txt";

	PresetsFormat = new TStringList();
  HeadersFormat = new TStringList();
  FootersFormat = new TStringList();
	LoadPresets();
  PageControl1->ActivePage = tabRAWData;  

  // TODO : do a better parameters parsing
  
	// Check if we have parameters
  int iParamCount = ParamCount();
	if( iParamCount > 0 )
	{
    // Check param range
    if(iParamCount != 1 && iParamCount != 3 && iParamCount != 4)
    {
        // Wrong number of parameters -> quit
  			AppliResult = 2;
  			PostQuitMessage(0);
  			return;
    }

		// Don't work in the last saved dir but in the apps dir
		SetCurrentDir(AppsPath);
		ifoFileName = ParamStr(1);

		// Check if ifo file exist
		if( !FileExists(ifoFileName) )
		{
      // error ifo file doesn't exist  -> quit
			AppliResult = 1;
			PostQuitMessage(0);
			return;
    }

    Caption = Application->Title + " - " + ifoFileName;
    AnsiString OutputFilename = ParamStr(2);

    parseIfo(true);
    if (ParamCount() != 1)
    {
      int p = -1, t = -1;
      if(ParamCount() == 4)
      {
        sscanf( ParamStr(4).c_str() , "-t%d", &t );
        // set the right title
        cbTitle->ItemIndex = t-1;
        parseIfo(false);
      } else {
        parseIfo(true);
      }

      if(iParamCount == 3 || iParamCount == 4)
      {
          sscanf( ParamStr(3).c_str() , "-p%d", &p );
          // set the right tab
          if(p > 0 && p <= 10)
          {
            PageControl1->ActivePage = tsFormat;
            cbPresets->ItemIndex = p-1;
            cbPresetsChange(this);
            CustomMemo->Lines->SaveToFile( OutputFilename );
          }
          else
          {
            Memo1->PlainText = true;
            Memo1->Lines->SaveToFile( OutputFilename );
          }
      }
      // OK
      AppliResult = 0;
      PostQuitMessage(0);
      return;
    }
  }
  
  cbPresets->ItemIndex = 0;
  cbPresetsChange(this);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::LoadPresets()
{
	TIniFile *fIni = new TIniFile( ChangeFileExt( Application->ExeName, ".INI" ) );

	for(int i = 0; i < 10; i++)
	{
		AnsiString name = "Preset " + AnsiString(i+1);
		AnsiString format = "Format " + AnsiString(i+1);
		AnsiString header = "Header " + AnsiString(i+1);
    AnsiString footer = "Footer " + AnsiString(i+1);

		cbPresets->Items->Add( fIni->ReadString("Presets", name, name) );
		PresetsFormat->Add( fIni->ReadString("Presets", format, format) );
		HeadersFormat->Add( fIni->ReadString("Presets", header, header) );
		FootersFormat->Add( fIni->ReadString("Presets", footer, footer) );
		SetCurrentDir(fIni->ReadString("Options", "LastDir", ""));
	}
 delete fIni;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::SavePresets()
{
	TIniFile *fIni = new TIniFile( ChangeFileExt( Application->ExeName, ".INI" ) );

	for(int i = 0; i < 10; i++)
	{
		AnsiString name = "Preset " + AnsiString(i+1);
		AnsiString format = "Format " + AnsiString(i+1);
		AnsiString header = "Header " + AnsiString(i+1);
    AnsiString footer = "Footer " + AnsiString(i+1);

		if(i < cbPresets->Items->Count)
		{
			fIni->WriteString("Presets", name, cbPresets->Items->Strings[i]);
			fIni->WriteString("Presets", format, PresetsFormat->Strings[i]);
			fIni->WriteString("Presets", header, HeadersFormat->Strings[i]);
			fIni->WriteString("Presets", footer, FootersFormat->Strings[i]);
			fIni->WriteString("Options", "LastDir", GetCurrentDir());
		}
	}
	delete fIni;
}
//---------------------------------------------------------------------------
AnsiString msToTime(unsigned long int i)
{
	char hh, min, sec, cent;

	hh = ceil(i / (3600 * 1000));
	min = ceil((i - (hh * 3600 * 1000)) / (60 * 1000));
	sec = ceil((i - (hh * 3600 * 1000) - (min * 60 * 1000)) / 1000);
	cent = ceil((i - (hh * 3600 * 1000) - (min * 60 * 1000) - (sec * 1000)) / 10);

	return AnsiString().sprintf("%02d:%02d:%02d:%02d", hh, min, sec, cent);
}
//---------------------------------------------------------------------------
#include <shlobj.h>

AnsiString _Tools_Browse_Root_Directory_ = "C:\\";
int __stdcall BrowseProc(HWND hwnd,UINT uMsg, LPARAM lParam, LPARAM lpData )
{
    char szDir[MAX_PATH];

    switch(uMsg)
    {
        case BFFM_INITIALIZED:
            // Set the initial directory. If WPARAM is TRUE, then LPARAM is a
            // string that contains the path. If WPARAM is FALSE, then LPARAM
            // should be a PIDL
            SendMessage(hwnd, BFFM_SETSELECTION, TRUE, (LPARAM)_Tools_Browse_Root_Directory_.c_str() );
            break;
    }
    return 0;
}

AnsiString BrowseForFolder(HWND WinHandle, AnsiString title, AnsiString root)
{
  PItemIDList IDRoot;
  TBrowseInfo BrowseInfo;
  PItemIDList PList;
  Char Path[MAX_PATH];

  _Tools_Browse_Root_Directory_ = ExcludeTrailingBackslash(root);
  AnsiString  FolderName;
  ZeroMemory(&BrowseInfo, sizeof(TBrowseInfo));
  ZeroMemory(Path, MAX_PATH);

  BrowseInfo.hwndOwner  =  WinHandle;
  BrowseInfo.pidlRoot   =  NULL;
  BrowseInfo.pszDisplayName = FolderName.c_str();
  BrowseInfo.lpszTitle  = title.c_str();
  BrowseInfo.lpfn       = NULL;
	BrowseInfo.ulFlags    = BIF_DONTGOBELOWDOMAIN | BIF_RETURNONLYFSDIRS | BIF_RETURNFSANCESTORS;
  BrowseInfo.lpfn       = BrowseProc;             // callback function


  PList = SHBrowseForFolder(&BrowseInfo);    //<-- show the dialog

  if (PList)
  {
    SHGetPathFromIDList(PList, Path);
		return IncludeTrailingBackslash( Path );
  } else
    return "";
}

//---------------------------------------------------------------------------
void __fastcall TMainForm::parseIfo(bool autoSelectPGCI)
{
	Memo1->Lines->Clear();
	Memo1->SelAttributes->Color = clBlack;	
	lstCells->Items->Clear();

	ifo = ifoOpen(ifoFileName.c_str(), NULL);
	if (ifo != NULL)
	{
		unsigned int i, j;

		Memo1->Lines->Add("Opening IFO file " + ifoFileName + " OK.");
		Memo1->Lines->Add("");

 		unsigned int numPGC = ifoGetNumPGCI(ifo);

		// Fill title combo box
		int savIdx = cbTitle->ItemIndex;

		cbTitle->Clear();
		for (i = 1; i <= numPGC; i++)
			cbTitle->Items->Add(i);

		if( autoSelectPGCI )
		{
			cbTitle->ItemIndex = SearchLongestTitle();
		} else {
			if (savIdx >= 0 && savIdx < numPGC)
				cbTitle->ItemIndex = savIdx;
			else
				cbTitle->ItemIndex = 0;
		}

		// Searching for the frame rate
		ifo_video_info_t *video = (ifo_video_info_t *) ifo->mat + IFO_OFFSET_VIDEO;

		// "PAL 625/50" or "NTSC 525/60"
		cbFrameRate->Text = AnsiString(video->system ? 25000 : 29970);

		// Print video info
		PrintVideoInfo();
		Memo1->Lines->Add("");

		// Print audio info
		j = ifoGetNumAudio(ifo);
		if (j > 0)
		{
			AddDecoratedLine("Audio:",fsBold);
		}
		for (i = 0; i < j; i++)
			Memo1->Lines->Add(AnsiString().sprintf("  %u. %s", i,ifoGetAudioDesc(ifo, i)));
		Memo1->Lines->Add("");

		// Print Subpicture info
		j = ifoGetNumSubPic(ifo);
		if (j > 0)
			AddDecoratedLine("SubPicture:",fsBold);

		for (i = 0; i < j; i++)
			Memo1->Lines->Add(AnsiString().sprintf("  %u. %s", i, ifoGetSubPicDesc(ifo, i)));

		Memo1->Lines->Add("");
		AddDecoratedLine("Program Chain(s):",fsBold);

		int ifo_pgc_idx = cbTitle->ItemIndex;
		dword num_cells;
		dword num_links;
		t_vs_vobcellid *cell_ids;

		if (ifo_pgc_idx >= 0)
		{
			if (ifo_pgc_idx < numPGC)
			{
				byte time[4];

				num_cells = ifoGetPGCIInfo(ifo, ifo_pgc_idx, time);
				cell_ids = new t_vs_vobcellid[num_cells];
				// frame 2 ms
				double frate = cbFrameRate->Text.ToInt() / 1000.0;
				double f2ms = time[3] * (1 / frate) * 1000;

				time[3] = f2ms / 10;

				Memo1->Lines->Add(AnsiString().sprintf("  Using PGC #%u/%u. Length: %02u:%02u:%02u:%02u in %i cell(s).",
					ifo_pgc_idx, numPGC, time[0], time[1], time[2], time[3], num_cells));
				ifoGetPGCICells(ifo, ifo_pgc_idx, cell_ids);

				// Calculate Time stamps
				unsigned long cumulTimeMs = 0;
				unsigned int currentTimeMs = 0;

				cellCumulTime.clear();
				cellTime.clear();
				cellCumulTime.push_back(cumulTimeMs);

				// Show each cells info & calculate timestamps
				for (j = 0; j < num_cells; j++)
				{
					// frame 2 ms
					f2ms = cell_ids[j].time_out[3] * (1 / frate) * 1000;

					// Calculate cell time in ms
					currentTimeMs =
					cell_ids[j].time_out[0] * 3600 * 1000 +
					cell_ids[j].time_out[1] * 60 * 1000 +
					cell_ids[j].time_out[2] * 1000 + f2ms;

					// Transform ms 2 cents
					cell_ids[j].time_out[3] = f2ms / 10;

					cellTime.push_back(currentTimeMs);

					TListItem *li = lstCells->Items->Add();

					li->Caption = AnsiString().sprintf("%03u", j);
					li->SubItems->Add(AnsiString().sprintf("%02u", cell_ids[j].vob_id));
					li->SubItems->Add(AnsiString().sprintf("%02u", cell_ids[j].cell_id));
					li->SubItems->Add(AnsiString().sprintf("%02u:%02u:%02u:%02u",
						cell_ids[j].time_out[0], cell_ids[j].time_out[1],
						cell_ids[j].time_out[2], cell_ids[j].time_out[3]));

					if (cell_ids[j].angle != 0x11)
					{
						if (((cell_ids[j].angle) & 0xf) == 1)
						{
							// We choose angle 1
							li->SubItems->Add(msToTime(cumulTimeMs));
							Memo1->Lines->Add(AnsiString().sprintf("  %03u. VOB-ID: %02u/CELL-ID: %02u - Len: %02u:%02u:%02u:%02u - TS: %s (Angle %01u/%01u)",
								j, cell_ids[j].vob_id, cell_ids[j].cell_id, cell_ids[j].time_out[0],
								cell_ids[j].time_out[1], cell_ids[j].time_out[2],
								cell_ids[j].time_out[3], msToTime(cumulTimeMs).c_str(),
									(cell_ids[j].angle) & 0xf, (cell_ids[j].angle) >> 4));
						}
						else
						{
							li->SubItems->Add("-");
							Memo1->Lines->Add(AnsiString().sprintf("  %03u. VOB-ID: %02u/CELL-ID: %02u - Len: %02u:%02u:%02u:%02u - TS: --:--:--:-- (Angle %01u/%01u)",
								j, cell_ids[j].vob_id, cell_ids[j].cell_id, cell_ids[j].time_out[0],
								cell_ids[j].time_out[1], cell_ids[j].time_out[2],
								cell_ids[j].time_out[3], (cell_ids[j].angle) & 0xf,
								(cell_ids[j].angle) >> 4));
							currentTimeMs = 0;
						}
						li->SubItems->Add(AnsiString().sprintf("%02u", (cell_ids[j].angle & 0xf)));
					}
					else
					{
						li->SubItems->Add(msToTime(cumulTimeMs));
						Memo1->Lines->Add(AnsiString().sprintf("  %03u. VOB-ID: %02u/CELL-ID: %02u - Len: %02u:%02u:%02u:%02u - TS: %s",
							j, cell_ids[j].vob_id, cell_ids[j].cell_id,
							cell_ids[j].time_out[0], cell_ids[j].time_out[1],
							cell_ids[j].time_out[2], cell_ids[j].time_out[3],
							msToTime(cumulTimeMs).c_str()));
						li->SubItems->Add("");
					}
					cumulTimeMs += currentTimeMs;
					cellCumulTime.push_back(cumulTimeMs);
				}

				// Get cells links
				num_links = ifoGetProgramMapNum(ifo, ifo_pgc_idx);
				nbsChapters = num_links;
				if (num_links > 0)
				{
					pgc_program_map_t *cell_links = new pgc_program_map_t[num_cells];

					ifoGetProgramMap(ifo, ifo_pgc_idx, cell_links);
					// Transfert table to vector
					cellLinks.clear();
					for (j = 0; j < num_links; j++)
						cellLinks.push_back(cell_links[j]);
					delete[]cell_links;

					// Print cells links
					Memo1->Lines->Add("");
					AddDecoratedLine("Cells links :",fsBold);
					AnsiString links = AnsiString().sprintf("%03u", cellLinks[0] - 1);

					TListItem *li;

					// allocate all strings to listview
					for (j = 0; j < num_cells; j++)
					{
						li = lstCells->Items->Item[j];
						li->SubItems->Add("");
						li->SubItems->Add("");
					}

					li = lstCells->Items->Item[0];
					li->SubItems->Strings[COL_LINK] =	AnsiString().sprintf("%03u", cellLinks[1] - 1);
					li->SubItems->Strings[COL_CHAPT] = "1";
					li->Checked = true;

					for (j = 1; j < num_links; j++)
					{
						links += " -> ";
						if ((j % 10) == 0)
							links += "\r\n";
						links += AnsiString().sprintf("%03u", cellLinks[j] - 1);

						if (j + 1 < num_links)
						{
							li = lstCells->Items->Item[cellLinks[j] - 1];
							li->SubItems->Strings[COL_LINK] = AnsiString().sprintf("%03u", cellLinks[j + 1] - 1);
							li->Checked = true;
						}
						else
						{
							// this is the last link so check the linked cells
							li = lstCells->Items->Item[cellLinks[j] - 1];
							li->Checked = true;
						}
						li->SubItems->Strings[COL_CHAPT] = (j + 1);
					}
					Memo1->Lines->Add(links);
				}

				Memo1->Lines->Add("");
				AddDecoratedLine("Total film length : " + msToTime(cumulTimeMs), fsBold);
				lbLstTimeLen->Caption = msToTime(cumulTimeMs);
				lbLstNbsChapter->Caption = nbsChapters;
				Memo1->Lines->Add("");
				for (unsigned int i = 0; i < nbsChapters; i++)
				{
					Memo1->Lines->Add("Chapter " + AnsiString(i + 1) + " = " +
						msToTime(cellCumulTime[cellLinks[i] - 1]));
				}

				// Last chapter bug workaround
				// Check size of last chapter
				bool bugIsPresent = chkFixLCB->Checked && cellTime[cellLinks[nbsChapters-1] - 1] < 5000;
				if( bugIsPresent )
				{
					Memo1->SelAttributes->Color = clRed;
					Memo1->Lines->Add("Warning ! Last chapter length is less than 5 s and has been disabled.");
					Memo1->SelAttributes->Color = clBlack;
					// Disable corresponding cell
					lstCells->Items->Item[ cellLinks[nbsChapters-1] - 1 ]->Checked = false;
					nbsChapters--;
					lbLstNbsChapter->Caption = nbsChapters;
					bugIsPresent = true;
				}

				// Goto Memo1 end
				// Set caret's position at end
				if(PageControl1->ActivePageIndex == 0 && ParamCount() == 0)
				{
					Memo1->SetFocus();
					Memo1->SelLength = 0;
					Memo1->SelStart = Memo1->Text.Length();
				}
				else
				{
					UpdateCurrentPage();
				}

				bttOpenAVI->Enabled = true;
				// Here we can free data
				delete[]cell_ids;
			}
			else
			{
				ifoClose(ifo);
				Memo1->Lines->Add(AnsiString().sprintf("* Program-Chain (PGC) Index out of range (only %i PGC(s) available).", numPGC));
				return;
			}
		}
		ifoClose(ifo);
	}
	else
	{
		Memo1->Lines->Add("error opening file " + ifoFileName);
		bttOpenAVI->Enabled = false;
		return;
	}
}

//---------------------------------------------------------------------------
int __fastcall TMainForm::SearchLongestTitle()
{
	int result = 0;

	if (ifo != NULL)
	{
		unsigned int num = ifoGetNumPGCI(ifo);
		byte time[4];
		long max_time = 0;

		for(int i=0; i < num; i++)
		{
			long currentTime = 0;
			ifoGetPGCIInfo(ifo, i, time);

			currentTime = time[0] * 3600 * 1000 +
				time[1] * 60 * 1000 +
				time[2] * 1000 + time[3];
			if(currentTime > max_time)
			{
				result = i;
				max_time = currentTime;
			}
		}
	} else {
		result = -1;
	}
	
	return result;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::bttOpenIFOClick(TObject * Sender)
{
	OpenDialog1->Filter = "DVD IFO File|*.IFO";
  OpenDialog1->FileName = ifoFileName;
  
	if (OpenDialog1->Execute())
	{
		ifoFileName = OpenDialog1->FileName;
		Caption = Application->Title + " - " + ifoFileName;
		parseIfo(true);
	}
}

//---------------------------------------------------------------------------

void __fastcall TMainForm::CalculateCustomChapters()
{
  PAVIFILE pfile;
  HRESULT hr;
  PAVISTREAM pStream;

	bool UsingAVIFile = (cbSynchKF->Enabled && cbSynchKF->Checked);

  if (UsingAVIFile)
  {
    // Init the AVI part
    // The AVIFileInit function initializes the AVIFile library.
    AVIFileInit();

    // Opening the AVI file
    hr = AVIFileOpen(&pfile,	// returned file pointer
		     edAVIFileName->Text.c_str(),	// file name
		     OF_READ,	// mode to open file with
		     NULL);	// use handler determined from file reading...

    if (hr != AVIERR_OK)
    {
      ShowMessage("# AVIFileOpen : AVI source init error !");
      AVIFileExit();
      return;
    }
    else
    {
      hr = AVIFileGetStream(pfile, &pStream, streamtypeVIDEO, 0);
      if (hr != AVIERR_OK)
      {
				ShowMessage("# AVIFileGetStream : Can't get a video stream !");
				AVIFileRelease(pfile);
				AVIFileExit();
				return;
      }
		}
	}

	int frameOff = edFrameOff->Text.ToInt();
	frameRate = cbFrameRate->Text.ToInt();
	int msOff = edMsOff->Text.ToInt();

  #define TEXTFIX(__x__) ((__x__->Text == NULL) ? "" : __x__->Text.c_str())

  AnsiString allText = "";
  if( !edHeader->Text.IsEmpty() )
  {
    AnsiString tmp = edHeader->Text;
		tmp = StringReplace(tmp, "%dir", edFormatDirectory->Text, TReplaceFlags() << rfReplaceAll);
		tmp = StringReplace(tmp, "%file", edFormatFileName->Text, TReplaceFlags() << rfReplaceAll);
    tmp = StringReplace(tmp,"\\n","\r\n", TReplaceFlags() << rfReplaceAll);
		tmp = StringReplace(tmp, "%sp", AnsiString(" "), TReplaceFlags() << rfReplaceAll);
    allText += tmp;
  }

  for (unsigned int i = 0; i < nbsChapters; i++)
	{
		int currentFrameNumber = (ceil((cellCumulTime[cellLinks[i] - 1] / 1000.0) * (frameRate / 1000.0)) + frameOff);
    if (currentFrameNumber < 0)
      currentFrameNumber = 0;

    if (UsingAVIFile)
    {
			long prevKF = AVIStreamFindSample(pStream, currentFrameNumber,
				FIND_KEY | FIND_PREV);
			long nextKF = AVIStreamFindSample(pStream, currentFrameNumber,
				FIND_KEY | FIND_NEXT);
			long nearestKF = (currentFrameNumber - prevKF < nextKF - currentFrameNumber) ? prevKF : nextKF;
			currentFrameNumber = nearestKF;
		}

		long ms = cellCumulTime[cellLinks[i] - 1] + msOff;
		if (ms < 0)
			ms = 0;

		int hh, min, sec, cent, ff, mss, as;
    float asf;
		hh = ceil(ms / (3600 * 1000));
		min = ceil((ms - (hh * 3600 * 1000)) / (60 * 1000));
		sec = ceil((ms - (hh * 3600 * 1000) - (min * 60 * 1000)) / 1000);
		cent = ceil((ms - (hh * 3600 * 1000) - (min * 60 * 1000) - (sec * 1000)) / 10);
    ff = ceil((ms - (hh * 3600 * 1000) - (min * 60 * 1000) - (sec * 1000)) / (1000000.0/frameRate));
    mss = ms - (hh * 3600 * 1000) - (min * 60 * 1000) - (sec * 1000);
    as = ceil(ms / 1000.0);
    asf = ms / 1000.0;

		AnsiString tmp = edFormat->Text;
		// %hh -> hours
		tmp = StringReplace(tmp, "%hh", AnsiString().sprintf("%02d",hh), TReplaceFlags() << rfReplaceAll);
		// %mm -> minutes
		tmp = StringReplace(tmp, "%mm", AnsiString().sprintf("%02d",min), TReplaceFlags() << rfReplaceAll);
		// %ss -> seconds
		tmp = StringReplace(tmp, "%ss", AnsiString().sprintf("%02d",sec), TReplaceFlags() << rfReplaceAll);
		// %cc -> cents
		tmp = StringReplace(tmp, "%cc", AnsiString().sprintf("%02d",cent), TReplaceFlags() << rfReplaceAll);
		// %c -> chapter number
		tmp = StringReplace(tmp, "%c", AnsiString(i+1), TReplaceFlags() << rfReplaceAll);
		// %C -> chapter number (2 digits)
		tmp = StringReplace(tmp, "%C", AnsiString().sprintf("%02d",i+1), TReplaceFlags() << rfReplaceAll);
    // %ff -> frames
		tmp = StringReplace(tmp, "%ff", AnsiString().sprintf("%02d",ff), TReplaceFlags() << rfReplaceAll);
		// %f -> frames
		tmp = StringReplace(tmp, "%f", AnsiString(currentFrameNumber), TReplaceFlags() << rfReplaceAll);
    // \n -> \r\n
    tmp = StringReplace(tmp, "\\n", "\r\n", TReplaceFlags() << rfReplaceAll);
		// %ms -> milliseconds
		tmp = StringReplace(tmp, "%ms", AnsiString().sprintf("%03d",mss), TReplaceFlags() << rfReplaceAll);
    // %asf -> all seconds float
		tmp = StringReplace(tmp, "%asf", AnsiString().sprintf("%.3f",asf), TReplaceFlags() << rfReplaceAll);
		// %as -> all seconds
		tmp = StringReplace(tmp, "%as", AnsiString().sprintf("%d",as), TReplaceFlags() << rfReplaceAll);
		// %ams -> all milliseconds
		tmp = StringReplace(tmp, "%ams", AnsiString().sprintf("%d",ms), TReplaceFlags() << rfReplaceAll);
    // %sp -> force space
		tmp = StringReplace(tmp, "%sp", AnsiString(" "), TReplaceFlags() << rfReplaceAll);

    allText += tmp;
	}

  if( !edFooter->Text.IsEmpty() )
  {
    AnsiString tmp = edFooter->Text;
		tmp = StringReplace(tmp, "%dir", edFormatDirectory->Text, TReplaceFlags() << rfReplaceAll);
		tmp = StringReplace(tmp, "%file", edFormatFileName->Text, TReplaceFlags() << rfReplaceAll);
    tmp = StringReplace(tmp,"\\n","\r\n", TReplaceFlags() << rfReplaceAll);
		tmp = StringReplace(tmp, "%sp", AnsiString(" "), TReplaceFlags() << rfReplaceAll);
    allText += tmp;
  }

  CustomMemo->Text = allText;

	if (UsingAVIFile)
  {
    AVIFileRelease(pfile);
    AVIFileExit();
  }
}

//---------------------------------------------------------------------------
void __fastcall TMainForm::bttQuitClick(TObject * Sender)
{
	PostQuitMessage(0);
}

//---------------------------------------------------------------------------
void __fastcall TMainForm::PrintVideoInfo()
{
  ifo_video_info_t *video = (ifo_video_info_t *) ifo->mat + IFO_OFFSET_VIDEO;

	AddDecoratedLine("Video:",fsBold);
  Memo1->Lines->Add(AnsiString().
		    sprintf(" Compression : MPEG-%d", video->compression + 1));
  Memo1->Lines->Add(AnsiString().
		    sprintf(" TV system   : %s Hz",
			    video->system ? "PAL 625/50" : "NTSC 525/60"));
  switch (video->ratio)
  {
    case 0:
      Memo1->Lines->Add(AnsiString().sprintf(" Aspect ratio: 4:3"));
      break;
    case 3:
      Memo1->Lines->Add(AnsiString().sprintf(" Aspect ratio: 16:9"));
      break;
    default:
      Memo1->Lines->Add(AnsiString().sprintf(" Aspect ratio: undefined"));
      break;
  }

  char perm_displ[][23] = { "pan-scan & letterboxed", "pan-scan",
    "letterboxed", "not specified"
  };

  Memo1->Lines->Add(AnsiString(" Display Mode: ") +
		    perm_displ[video->perm_displ]);

  char source_res[][28] = { "720x480", "704x480", "352x480", "352x240",
    "720x576", "704x576", "352x576", "352x288"
  };

  Memo1->Lines->Add(AnsiString(" Source Res  : ") +
		    source_res[video->source_res + (video->system * 4)]);

  Memo1->Lines->Add(AnsiString(" Letterboxed : ") +
		    (video->letterboxed ? "yes" : "no"));
  Memo1->Lines->Add(AnsiString(" Mode        : ") +
		    (video->mode ? "Film (625/50 only)" : "Camera"));
}

//---------------------------------------------------------------------------
void __fastcall TMainForm::edFrameOffExit(TObject * Sender)
{
	TEdit* ed = (TEdit*)Sender;
	// Check if Frame offset entry is OK
	if (!ed->Text.IsEmpty())
  {
    try
    {
			ed->Text.ToInt();
    }
    catch(EConvertError & e)
    {
			ed->Text = "0";
    }
  }
  else
    ed->Text = "0";
}

//---------------------------------------------------------------------------
void __fastcall TMainForm::bttHelpAboutClick(TObject * Sender)
{
  // Show about box
  FormAbout->Timer1->Interval = 6000;
  FormAbout->ShowModal();
}

//---------------------------------------------------------------------------
void __fastcall TMainForm::FormActivate(TObject * Sender)
{
  static bool firstTime = true;

  if(ParamCount() != 0)
  {
    firstTime = false;
  }
  if (firstTime)
  {
    FormAbout->Timer1->Interval = 2000;
    FormAbout->ShowModal();
    firstTime = false;
  }

}

//---------------------------------------------------------------------------

void __fastcall TMainForm::cbFrameRateExit(TObject * Sender)
{
  // Check if Frame rate entry is OK
  if (!cbFrameRate->Text.IsEmpty())
  {
    try
    {
      cbFrameRate->Text.ToInt();
    }
    catch(EConvertError & e)
    {
      cbFrameRate->Text = "25000";
    }
  }
  else
    cbFrameRate->Text = "25000";
}

//---------------------------------------------------------------------------

void __fastcall TMainForm::bttOpenAVIClick(TObject * Sender)
{
  OpenDialog1->Filter = "AVI Video File|*.AVI";
  OpenDialog1->FileName = aviFileName;
  
  if (OpenDialog1->Execute())
  {
    aviFileName = OpenDialog1->FileName;
    edAVIFileName->Text = OpenDialog1->FileName;
    // Init the AVI part
    // The AVIFileInit function initializes the AVIFile library.
    AVIFileInit();

    // Opening the AVI file
    PAVIFILE pfile;
    HRESULT hr;

    hr = AVIFileOpen(&pfile,	// returned file pointer
		     OpenDialog1->FileName.c_str(),	// file name
		     OF_READ,	// mode to open file with
		     NULL);	// use handler determined from file reading...

    if (hr != AVIERR_OK)
    {
      ShowMessage("# AVIFileOpen : AVI source init error !");
      cbSynchKF->Enabled = false;
    }
    else
    {
      // Retrieving AVI file FPS
      AVIFILEINFO fi;

      AVIFileInfo(pfile, &fi, sizeof(AVIFILEINFO));
			cbFrameRate->Text = AnsiString(ceil((((double) fi.dwRate) / fi.dwScale) * 1000));
      cbSynchKF->Enabled = true;
      AVIFileRelease(pfile);
		}

		UpdateCurrentPage();
		
    AVIFileExit();
  }

}

//---------------------------------------------------------------------------
void __fastcall TMainForm::cbTitleChange(TObject * Sender)
{
  parseIfo(false);
}

//---------------------------------------------------------------------------
void __fastcall TMainForm::lstCellsMouseUp(TObject * Sender, TMouseButton Button,
					TShiftState Shift, int X, int Y)
{
	THitTests ht = lstCells->GetHitTestInfoAt(X,Y);

	if (ht.Contains(htOnStateIcon))
	{
		UpdateListLinkAndChapters();
	}
}

//---------------------------------------------------------------------------

void __fastcall TMainForm::lstCellsCustomDrawItem(TCustomListView * Sender,
					       TListItem * Item,
					       TCustomDrawState State,
					       bool & DefaultDraw)
{
  if (Item != NULL && !State.Contains(cdsHot))
  {
    if (Item->SubItems->Strings[COL_TS] == "-")
    {
      // If timestamp == "-", show it in grey
      lstCells->Canvas->Font->Color = clGrayText;
    }
    else if (Item->Checked)
    {
      // Show chapters in red        
      lstCells->Canvas->Font->Color = clRed;
    }
  }
}

//---------------------------------------------------------------------------
void __fastcall TMainForm::UpdateListLinkAndChapters()
{
	// Update link list
	cellLinks.clear();
	bool isFirst = true;
	for (int i = 0; i < lstCells->Items->Count; i++)
	{
		if (lstCells->Items->Item[i]->Checked)
		{
			if(isFirst)
			{
					cellLinks.push_back(i + 1);
					isFirst = false;
			}
			// Search next checked
			for (int j = i + 1; j < lstCells->Items->Count; j++)
			{
				if (lstCells->Items->Item[j]->Checked)
				{
					cellLinks.push_back(j + 1);
					break;
				}
			}
		}
	}

	nbsChapters = cellLinks.size();
	lbLstNbsChapter->Caption = nbsChapters;

	// Update list view
	TListItem *li;

	// Clear all links & chapters
	for (int i = 0; i < lstCells->Items->Count; i++)
	{
		li = lstCells->Items->Item[i];
		li->SubItems->Strings[COL_LINK] = "";
		li->SubItems->Strings[COL_CHAPT] = "";
	}

	// Now rewrite right ones
	for (unsigned int i = 0; i < nbsChapters; i++)
	{
		if (i + 1 < nbsChapters)
		{
			li = lstCells->Items->Item[cellLinks[i] - 1];
			li->SubItems->Strings[COL_LINK] =	AnsiString().sprintf("%03u", cellLinks[i + 1] - 1);
		}
		li->SubItems->Strings[COL_CHAPT] = (i + 1);
	}
}


//---------------------------------------------------------------------------
void __fastcall TMainForm::RecalculateTSFromLst()
{
  unsigned long cumulTimeMs = 0;
	TListItem *li;

  cellCumulTime.clear();
  cellCumulTime.push_back(0);

  // Update vector
  for (int i = 0; i < lstCells->Items->Count; i++)
  {
    li = lstCells->Items->Item[i];
    if (li->SubItems->Strings[COL_TS] != "-")
    {
      cumulTimeMs += cellTime[i];
      cellCumulTime.push_back(cumulTimeMs);
    }
    else
		{
      cellCumulTime.push_back(cumulTimeMs);
    }
	}

	// Update listview
	for (int i = 0; i < lstCells->Items->Count; i++)
	{
		li = lstCells->Items->Item[i];
		if (li->SubItems->Strings[COL_TS] != "-")
			li->SubItems->Strings[COL_TS] = msToTime(cellCumulTime[i]);
	}
	lbLstTimeLen->Caption = msToTime(cumulTimeMs);
}

//---------------------------------------------------------------------------

void __fastcall TMainForm::Useintimestamp1Click(TObject * Sender)
{
	ChangeSelected("x",false);
	// Recalculate TimeStamp
	RecalculateTSFromLst();
}

//---------------------------------------------------------------------------

void __fastcall TMainForm::Dontuseintimestamp1Click(TObject * Sender)
{
	ChangeSelected("-",false);
	// Recalculate TimeStamp
	RecalculateTSFromLst();
}

//---------------------------------------------------------------------------

void __fastcall TMainForm::ChangeSelected(AnsiString s, bool checked)
{
	TListItem *li = lstCells->Selected;

	if (li != NULL)
	{
		int startAt = li->Index;
		for(int i = startAt; i < lstCells->Items->Count; i++)
		{
			li = lstCells->Items->Item[i];
			if( li->Selected )
			{
				if( s != NULL)
					li->SubItems->Strings[COL_TS] = s;
				else
					li->Checked = checked;
			}
		}
	}
}

//---------------------------------------------------------------------------
void __fastcall TMainForm::UpdateCurrentPage()
{
  switch(PageControl1->ActivePageIndex)
  {
    case 0:
      bttSaveData->Enabled = true;
      break;
    case 1:
  		// Update column size in case of scroll bar appear
  		// TODO : find a better way to do that ;-)
  		lstCells->Visible = false;
  		lstCells->Width = lstCells->Width + 1;
  		lstCells->Width = lstCells->Width - 1;
  		lstCells->Visible = true;
      bttSaveData->Enabled = false;
      break;
    case 2:
      bttSaveData->Enabled = true;
  		CalculateCustomChapters();
      break;
    case 3:
      bttSaveData->Enabled = false;    
      break;
  }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::PageControl1Change(TObject * Sender)
{
	UpdateCurrentPage();
}

//---------------------------------------------------------------------------
void __fastcall TMainForm::cbFrameRateChange(TObject * Sender)
{
	UpdateCurrentPage();
}

//---------------------------------------------------------------------------
void __fastcall TMainForm::edFrameOffKeyPress(TObject * Sender, char &Key)
{
	if (Key == VK_RETURN)
	{
		UpdateCurrentPage();
	}
	else if (!(isdigit(Key) || Key == '-' || Key == VK_BACK) || Key < 0)
	{
		Key = NULL;
	}
}

//---------------------------------------------------------------------------

void __fastcall TMainForm::AddDecoratedLine(AnsiString text, TFontStyle fs)
{
	TFontStyles oldStyle = Memo1->SelAttributes->Style;

	Memo1->SelAttributes->Style = Memo1->SelAttributes->Style << fs;
	Memo1->Lines->Add(text);
	Memo1->SelAttributes->Style = oldStyle;
}

//---------------------------------------------------------------------------

void __fastcall TMainForm::AddDecoratedLine2(AnsiString text)
{
	// TODO : translate 'html' like ansistring in rtf and insert it in RichEdit
	// <b></b> - bold
	// <i></i> - italic
	// <u></u> - underline
	// <strike></strike> - strike out
	// <FONT COLOR="#000000"  SIZE="12"> - rgb color  font size
}

//---------------------------------------------------------------------------

void __fastcall TMainForm::pmLstCellsPopup(TObject *Sender)
{
	// check & disable pop up menu
	bool sel = (lstCells->Selected != NULL);

	Useintimestamp1->Enabled = sel;
	Dontuseintimestamp1->Enabled = sel;
	Uncheck1->Enabled = sel;
	Check1->Enabled = sel;	
}

//---------------------------------------------------------------------------

void __fastcall TMainForm::Uncheck1Click(TObject *Sender)
{
	ChangeSelected(NULL,false);
	UpdateListLinkAndChapters();	
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::Check1Click(TObject *Sender)
{
	ChangeSelected(NULL,true);
	UpdateListLinkAndChapters();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::cbSynchKFClick(TObject *Sender)
{
	UpdateCurrentPage();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormDestroy(TObject *Sender)
{
	SavePresets();
	delete PresetsFormat;
  delete HeadersFormat;
  delete FootersFormat;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::bttModifPresetNameClick(TObject *Sender)
{
	if(cbPresets->Style == csSimple)
	{
		cbPresets->Items->Strings[cbPresets->Tag] = cbPresets->Text;
		cbPresets->Style = csDropDownList;
		cbPresets->ItemIndex = cbPresets->Tag;
	} else {
		if(cbPresets->ItemIndex != -1)
		{
			cbPresets->Tag = cbPresets->ItemIndex;
			cbPresets->Style = csSimple;
			cbPresets->SetFocus();
		}
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::cbPresetsKeyDown(TObject *Sender, WORD &Key,
			TShiftState Shift)
{
	if(cbPresets->Style == csSimple && Key == VK_RETURN)
	{
		cbPresets->Items->Strings[cbPresets->Tag] = cbPresets->Text;
		cbPresets->Style = csDropDownList;
		cbPresets->ItemIndex = cbPresets->Tag;
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::bttSavePresetFormatClick(TObject *Sender)
{
	if(cbPresets->ItemIndex != -1)
  {
		PresetsFormat->Strings[cbPresets->ItemIndex] = edFormat->Text;
    HeadersFormat->Strings[cbPresets->ItemIndex] = edHeader->Text;
    FootersFormat->Strings[cbPresets->ItemIndex] = edFooter->Text;
  }
	CalculateCustomChapters();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::cbPresetsChange(TObject *Sender)
{
	if(cbPresets->ItemIndex != -1)
  {
		edFormat->Text = PresetsFormat->Strings[cbPresets->ItemIndex];
		edHeader->Text = HeadersFormat->Strings[cbPresets->ItemIndex];
		edFooter->Text = FootersFormat->Strings[cbPresets->ItemIndex];
  }

	if(cbPresets->Style == csDropDownList)
		CalculateCustomChapters();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::bttSaveDataClick(TObject *Sender)
{
  if(SaveDialog1->Execute())
  {
    if(PageControl1->ActivePageIndex == 0)
    {
      Memo1->PlainText = true;
  		Memo1->Lines->SaveToFile( SaveDialog1->FileName );
    } else if(PageControl1->ActivePageIndex == 2) {
			CustomMemo->Lines->SaveToFile( SaveDialog1->FileName );
    }
  }
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::bttBrowseFilenameClick(TObject *Sender)
{
	OpenDialog1->Filter = "AC3 Files|*.AC3|All files|*.*";
  OpenDialog1->FileName = FormatFileName;

  if (OpenDialog1->Execute())
  {
    FormatFileName = OpenDialog1->FileName;
    edFormatFileName->Text = OpenDialog1->FileName;
    if( edFormatDirectory->Text.IsEmpty() )
      edFormatDirectory->Text = IncludeTrailingBackslash(ExtractFileDir(OpenDialog1->FileName));
  }
  CalculateCustomChapters();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::bttBrowseDirClick(TObject *Sender)
{
  edFormatDirectory->Text = BrowseForFolder(Handle, "Select a directory : ",
    edFormatDirectory->Text.c_str());
  CalculateCustomChapters();
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::bttGOClick(TObject *Sender)
{
  BOOL result;
  PROCESS_INFORMATION ProcessInfo;
  STARTUPINFO StartupInfo = {0};
  StartupInfo.cb = sizeof(STARTUPINFO);
  StartupInfo.dwFlags=STARTF_USESHOWWINDOW;
  StartupInfo.wShowWindow=SW_SHOWNORMAL;

   if( FileExists(AppsPath + "logfile.txt") )
   {
    DeleteFile(AppsPath + "logfile.txt");
   }

  result = CreateProcess(NULL, CustomMemo->Text.c_str(), NULL, NULL, FALSE,0, NULL,
    AppsPath.c_str(), &StartupInfo, &ProcessInfo);

  if(result)
  {
    while(WaitForSingleObject(ProcessInfo.hProcess, 100)!=WAIT_OBJECT_0)
    {
      MainForm->Refresh();
    }
    if( FileExists(AppsPath + "logfile.txt") )
    {
      ShellExecute(Handle, "open", "logfile.txt", NULL, AppsPath.c_str(), SW_SHOWNORMAL);
    }
  } else {
    Application->MessageBox(
      AnsiString().sprintf("Command %s failed ! Check that the executable is in ChapterXtractor's directory.",
      Trim(CustomMemo->Text).c_str()).c_str(),
      NULL, MB_OK|MB_ICONERROR);
  }
}
//---------------------------------------------------------------------------

